/* CA server:  sign requests, revoke certs, issue crls */

#include "caserver.h"

int verbose = 0;
char *prog;

/* unrecoverable errors. Mostly on startup. */

static void handle_error(const char *file, int lineno, const char *msg)
{
    fprintf(stderr, "** %s:%i %s\n", file, lineno, msg);
    ERR_print_errors_fp(stderr);
    exit(-1);
}

#define ERR(msg)  handle_error(__FILE__, __LINE__, msg)

/* warning messages - bad requests */

static void warn(const char *msg)
{
    fprintf(stderr, " --  %s\n", msg);
    ERR_print_errors_fp(stderr);
}

/* --------- ssl connection tools -------------------- */

/* this callback just reports the error */
int verify_callback(int ok, X509_STORE_CTX *store)
{
   char data[256];
 
   if (!ok) {
     X509 *cert = X509_STORE_CTX_get_current_cert(store);
     int  depth = X509_STORE_CTX_get_error_depth(store);
     int  err = X509_STORE_CTX_get_error(store);
 
     fprintf(stderr, "-Error with certificate at depth: %i\n", depth);
     X509_NAME_oneline(X509_get_issuer_name(cert), data, 256);
     fprintf(stderr, "  issuer   = %s\n", data);
     X509_NAME_oneline(X509_get_subject_name(cert), data, 256);
     fprintf(stderr, "  subject  = %s\n", data);
     fprintf(stderr, "  err %i:%s\n", err, X509_verify_cert_error_string(err));
    }
    return ok;
}

SSL_CTX *setup_server_ctx(void)
{
    SSL_CTX *ctx;
 
    ctx = SSL_CTX_new(SSLv23_method(  ));
    if (SSL_CTX_load_verify_locations(ctx, CA_CERT, NULL) != 1) ERR(CA_CERT);
    if (SSL_CTX_set_default_verify_paths(ctx) != 1) ERR("def CA");
    if (SSL_CTX_use_certificate_chain_file(ctx, SV_CERT) != 1) ERR(SV_CERT);
    if (SSL_CTX_use_PrivateKey_file(ctx, SV_CERT, SSL_FILETYPE_PEM) != 1)
                 ERR(SV_CERT);
    SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT,
                 verify_callback);
    SSL_CTX_set_verify_depth(ctx, 1);
    return ctx;
}
 

void init_OpenSSL(void)
{
    if (!THREAD_setup() || !SSL_library_init()) ERR("OpenSSL init");
    SSL_load_error_strings();
}


/* Check the cert's cn */
long cert_cn_check(SSL *ssl, char *cn)
{
    X509      *cert;
    X509_NAME *subj;
    char      data[256];
    int       extcount;
    int       ok = 0;
 
    if (!(cert = SSL_get_peer_certificate(ssl)) || !cn) goto err;

    if (!ok && (subj = X509_get_subject_name(cert)) &&
        X509_NAME_get_text_by_NID(subj, NID_commonName, data, 256) > 0) {
       data[255] = 0;
       if (strcasecmp(data, cn) != 0) goto err;
    }
 
    X509_free(cert);
    return SSL_get_verify_result(ssl);
 
err:
    if (cert) X509_free(cert);
    return X509_V_ERR_APPLICATION_VERIFICATION;
}


void seed_prng(void)
{
  struct timeval tv;
  gettimeofday(&tv,NULL);
  RAND_seed(&tv.tv_usec, 4);
  RAND_load_file(CA_ROOT "etc/rand", 1024);
  gettimeofday(&tv,NULL);
  RAND_seed(&tv.tv_usec, 4);
}



int process_request(SSL *ssl)
{
   int  err;
   CmdHdr_ cmdhdr;
   RspHdr_ rsphdr;
   int msgl = 0;
   char *camsg = NULL;
   char *caresp;
   int castatus;
   int rspl;
 
if (verbose) fprintf(stderr, "reading %d header bytes\n", sizeof(CmdHdr_));
   err = SSL_read(ssl, &cmdhdr, sizeof(CmdHdr_));
   if (err <= 0) {
fprintf(stderr, "reading error %d: %d\n", err, SSL_get_error(ssl, err));
       return (SSL_get_shutdown(ssl)&SSL_RECEIVED_SHUTDOWN)?1:0;
   }

   cmdhdr.cmd = ntohl(cmdhdr.cmd);
   cmdhdr.ext = ntohl(cmdhdr.ext);
   cmdhdr.flags = ntohl(cmdhdr.flags);
   msgl = ntohl(cmdhdr.len);
   if (msgl>=CA_MAXARG) {
      return(7);
   }
    
if (verbose) fprintf(stderr, "reading %d aux bytes\n", msgl);
   camsg = (char*) malloc(msgl+1);
   err = SSL_read(ssl, camsg, msgl);
   if (err <= 0) {
      free(camsg);
fprintf(stderr, "reading error %d: %d\n", err, SSL_get_error(ssl, err));
      return (SSL_get_shutdown(ssl)&SSL_RECEIVED_SHUTDOWN)?2:0;
   }
   camsg[msgl] = '\0';

   /* process the request's command */

   caresp = NULL;

   castatus = ca_cmd(&cmdhdr, camsg, &rsphdr, &caresp);

   free(camsg);

   /* send response */

   rsphdr.rsp = htonl(castatus);
   if (caresp) rspl = strlen(caresp);
   else rspl = 0;
   if (rspl) rsphdr.len = htonl(rspl);
   else rsphdr.len = 0;

if (verbose) fprintf(stderr, "writing %d header bytes\n", sizeof(RspHdr_));
   err = SSL_write(ssl, &rsphdr, sizeof(RspHdr_));
   if (err <= 0) {
       if (caresp) free(caresp);
       return (21);
   }
   if (rspl) {
if (verbose) fprintf(stderr, "writing %d response bytes\n", rspl);
      err = SSL_write(ssl, caresp, rspl);
      if (err <= 0) {
          if (caresp) free(caresp);
          return (22);
      }
   }

   if (caresp) free(caresp);

   return ( (SSL_get_shutdown(ssl)&SSL_RECEIVED_SHUTDOWN)? 3:0);
}



/* Thread to handle a request */

void *THREAD_CC server_thread(void *arg)
{
    SSL *ssl = (SSL *)arg;
    int err;
 
    pthread_detach(pthread_self());

    if (SSL_accept(ssl) <= 0) {
       fprintf(stderr, "Error: invalid connection.");
       sleep(1);
       return;
    }

    if ((err = cert_cn_check(ssl, CL_CN)) != X509_V_OK)
    {
        fprintf(stderr, "Error: peer certificate: %s\n",
                X509_verify_cert_error_string(err));
        fprintf(stderr, "  .. no connection\n");
        SSL_shutdown(ssl);
        goto err;
    }

    if (verbose) fprintf(stderr, "Connection opened\n");
    if (process_request(ssl)) SSL_shutdown(ssl);
    else SSL_clear(ssl);
    if (verbose) fprintf(stderr, "Connection closed\n");
err:
    SSL_free(ssl);
    ERR_remove_state(0);
}

char *prog;
static void usage()
{
  fprintf(stderr,"usage: %s [-p port] \n", prog);
  exit (1);
}

int main(int argc, char *argv[])
{
    BIO     *acc, *client;
    SSL     *ssl;
    SSL_CTX *ctx;
    THREAD_TYPE tid;
    time_t now;
    int acpt;
    FILE *pidf;
    char *caport = CA_PORT;

    prog = argv[0];
    while (--argc) {
       argv++;
       if (!strcmp(argv[0], "-p")) {
          if (--argc<0) usage();
          caport = (++argv)[0];
          printf(".. using port %s\n", caport);
       } else if (!strcmp(argv[0], "-v")) {
          verbose = 1;
          printf(".. verbose\n");
       } else usage();
    }

    init_OpenSSL(  );
    seed_prng(  );
 
    ctx = setup_server_ctx(  );
 
    acc = BIO_new_accept(caport);
    if (!acc) ERR("Error creating server socket");
 
    if (BIO_do_accept(acc) <= 0) ERR("Error binding server socket");
 
    pidf = fopen(CA_PID,"w");
    if (pidf) {
       fprintf(pidf,"%d\n", getpid());
       fclose(pidf);
    }

    for (;;) {
        acpt = BIO_do_accept(acc);
        now = time(NULL);
        fprintf(stderr, "\nAccept\n%s", ctime(&now));
        if (acpt <= 0) {
           warn("accept error");
           sleep(1);
           continue;
        }
        client = BIO_pop(acc);
        if (!(ssl = SSL_new(ctx))) {
           warn("create context error");
           sleep(1);
           continue;
        }
        SSL_set_accept_state(ssl);
        SSL_set_bio(ssl, client, client);
        THREAD_CREATE(tid, server_thread, ssl);
    }
 
    SSL_CTX_free(ctx);
    BIO_free(acc);
    exit (0);
}

