#include <stdio.h>
#include <time.h>
#include <netdb.h>
#include <resolv.h>
#include <string.h>

#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/ssl.h>
#include <openssl/buffer.h>
#include <openssl/objects.h>
#include <openssl/evp.h>
#include <openssl/rand.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#include <openssl/pem.h>
#include <openssl/pkcs12.h>
#include <openssl/des.h>

#include "cassl.h"

/** Certificate requests we will allow.

    If 'req' then the element must be present.
    If 'text' then:  
                   if '-dns-' we must have a valid dns name
                   else the text must match

 **/

typedef struct AllowDN__ {
  int nid;
  int req;
  char *text;
  char chk;
} AllowDN_, *AllowDN;

/* Server certs */
AllowDN_ serverdn[] = {
 { NID_countryName, 0, "US" ,0 },
 { NID_stateOrProvinceName, 0, "WA" ,0 },
 { NID_stateOrProvinceName, 0, "Washington" ,0 },
 { NID_localityName, 0, NULL ,0 },
 { NID_organizationName, 0, NULL ,0 },
 { NID_organizationalUnitName, 0, NULL ,0 },
 { NID_organizationalUnitName, 0, NULL ,0 },
 { NID_pkcs9_emailAddress, 0, NULL ,0 },
 { NID_commonName, 1, "-dns-" ,0 },
 { NID_undef, 0, NULL ,0 }
};

/* EDS certs ( cn=domain,ou=apps,dc=washington.dc=edu ) */
AllowDN_ edsdn[] = {
 { NID_domainComponent, 1, "edu" ,0 },
 { NID_domainComponent, 1, "washington" ,0 },
 { NID_organizationalUnitName, 1, "apps" ,0 },
 { NID_commonName, 1, "-dns-" ,0 },
 { NID_undef, 0, NULL ,0 }
};

static AllowDN allowdn[] = { serverdn, edsdn, NULL };

static void reset_adn(AllowDN a)
{
  while (a->nid!=NID_undef) {
     a->chk = 0;
     a++;
  }
}

/* Make sure no special chars in dns name */

void fix_dns_name(char *n)
{
   for (;*n;n++) {
      if (isupper(*n)) *n = tolower(*n);
      if (isalnum(*n)||(*n=='.')||(*n=='-')||(*n=='*')) continue;
      *n = '_';
   }
}


/* Verify that the user is an owner of the cn.
   Return NULL on OK, else return an error message. */

char *ssl_chk_netacts(char *cn, int lcn, char *id) {
  char *hdom;
  char *msg = NULL;
  int r;
  if (lcn==0) lcn = strlen(cn);
  hdom = (char*) malloc(lcn+1);
  memcpy(hdom,cn,lcn);
  hdom[lcn] = '\0';
  fix_dns_name(hdom);
  r = netact_test_host(hdom, id);
  if (!r) {
     char *edp, *dp;
     /* maybe owns the domain */
     if (edp=strrchr(hdom,'.')) {
        for (dp=strchr(hdom,'.');dp&&dp<edp;dp=strchr(dp+1,'.')) {
          if (r=netact_test_domain(dp+1, id)) break;
        }
     }
     if (!r) {
       msg = (char*) malloc(lcn+64);
       sprintf(msg,"CN: %s is not an owner of the host '%s'", id, hdom);
     }
  }
#ifdef NOWILDCARDS
  if ((!msg) && strchr(hdom,'*')) {
     msg = (char*) malloc(lcn+48);
     sprintf(msg,"WC: wildcard domains require administrator approval");
  }
#endif
  free(hdom);
  return (msg);
}


/* Verify that a cn is a valid DNS name.
   Accept any record.
   Allow "*." wildcard.
   Return NULL on OK, else return an error message. */

char *ssl_chk_dns(char *cn, int lcn) {
  char *msg = NULL;
  char rbuf[PACKETSZ];
  char *hdom;
  int r;

  if (lcn==0) lcn = strlen(cn);
  if (lcn<3 || strchr(cn, '<')) {
    msg = (char*) malloc(lcn+48);
    sprintf(msg,"DNS domain '%s' is invalid", cn);
  }
     
// don't bother.  netops doesn't seem to actually create all dns names
  if (1) return (NULL);

  hdom = (char*) malloc(lcn+2);
  if (strncmp(cn,"*.",2)) memcpy(hdom,cn,lcn);
  else {
    lcn -= 2;
    memcpy(hdom,cn+2,lcn);
  }
  hdom[lcn] = '.';
  hdom[lcn+1] = '\0';
  fix_dns_name(hdom);
  res_init();
  r = res_search(hdom, C_IN, T_ANY, rbuf, PACKETSZ);
  if (r<0) {
    msg = (char*) malloc(lcn+48);
    sprintf(msg,"DNS domain '%s' was not found", hdom);
  }
  free(hdom);
  return (msg);
}



/* Verify a requests against the list of request policies.
   Return NULL on success.
   Return error string on failure request failed. */

char *ssl_chk_req_dn(X509_REQ *req, char *id)
{
  X509_NAME *subj;
  X509_NAME_ENTRY *ent;
  BIO *cbio;
  int ni;
  int i;
  char txt[256];
  int ret;
  AllowDN a, adns;
  int j;
  char *emsg = NULL;
  
  subj = X509_REQ_get_subject_name(req);

  ni = X509_NAME_entry_count(subj);

  /* compare against each set */
  for (i=0; allowdn[i]; i++) {
    adns = allowdn[i];
    reset_adn(adns);
    /* check each element in the dn */
    for (j=0;j<ni;j++) {
       int nid;
       ent = X509_NAME_get_entry(subj, j);
       nid = OBJ_obj2nid(ent->object);
       for (a=adns;a->nid!=NID_undef;a++) {
          if (a->chk) continue;
          if (a->nid!=nid) continue;
          if (id && a->text && !strcmp(a->text,"-dns-")) {
            if (strcmp(id,"uwit")) {   // uwit is cert services website id
               emsg = ssl_chk_netacts(ent->value->data, ent->value->length, id);
               if (emsg) return (emsg);
            }
          } else {
            if (a->text && strcasecmp(a->text,ent->value->data)) continue; 
          }
          a->chk = 1; /* OK so far */
          break;
       }
       if (a->nid==NID_undef) break;
    }
    /* check for missing elements */
    for (a=adns;a->nid!=NID_undef;a++) if (a->req && !a->chk) break;
    if (a->nid==NID_undef) return (NULL); /* OK */
  }

  /* no match - reject */
  return("DN: The distinguished name does not match our policy");
}



/* Verify a request's altnames for ownership 
   Return NULL on success.
   Return error strign on failure. */

char *ssl_chk_req_altnames(X509_REQ *req, char *id)
{
   char **names = ssl_get_altnames_req(req);
   char *emsg = NULL;
   if (!id) return;

   for (names=ssl_get_altnames_req(req);names && *names; names++) {
      if (strcmp(id,"uwit")) {   // uwit is cert services website id
         emsg = ssl_chk_netacts(*names, 0, id);
         if (emsg) return (emsg);
      }
   }
   /* should free names */
   return (NULL);
}


/* Verify a request's altnames for DNS validity 
   Return NULL on success.
   Return error string on failure. */

char *ssl_chk_req_alt_dns(X509_REQ *req)
{
   char **names = ssl_get_altnames_req(req);
   char *emsg;

   for (names=ssl_get_altnames_req(req);names && *names; names++) {
      emsg = ssl_chk_dns(*names, 0);
      if (emsg) return (emsg);
   }
   /* should free names */
   return (NULL);
}
      
 
