/* ========================================================================
 * Copyright (c) 2004-2006 The University of Washington
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ========================================================================
 */


/* EZplot plotting library */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include <gd.h>
#include <gdfontt.h>
#include <gdfonts.h>
#include <gdfontmb.h>
#include <gdfontl.h>
#include <gdfontg.h>

#define EZP_LIBRARY

#include "ezp.h"
#include "rgb.h"


int ezp_verbose = 0;
#define PRINTF if (ezp_verbose) printf

#define Y(y) (I->height-(y))

#define UC (unsigned char *)

/* some time ranges in seconds */

#define MIN   60
#define HOUR  3600
#define DAY   86400
#define MON   86400*28
#define YEAR   86400*365

/* divide year ticks into 4 pieces (should compute) */
#define NYST  4

static double infinity;

/* Make a text string */

EZP_Text ezp_text(char *str, gdFontPtr font, int ci)
{
   EZP_Text T = (EZP_Text) malloc(sizeof(EZP_Text_));
   T->str = strdup(str);
   T->font = font;
   T->ci = ci;
   T->height = T->font->h;
   T->width = T->font->w * strlen(str);
   T->nl = 1;
   while (str=strchr(str,'\n')) str++,T->nl++;
   return (T);
}


int ezp_set_transparent_bg(EZP_Image I)
{
   gdImageColorTransparent(I->gdimg, 0);
}

EZP_Image ezp_create_image(int width, int height)
{
   EZP_Image I = (EZP_Image) malloc(sizeof(EZP_Image_));
   memset(I, 0, sizeof(EZP_Image_));
   I->gdimg = gdImageCreate(width, height);
   I->graph = NULL;
   I->ngraph = 0;
   I->title = NULL;
   I->width = width;
   I->height = height;
   I->legend = NULL;
   I->isplotted = 0;
   return(I);
}


void ezp_image_title(EZP_Image I, char *title, EZP_Color color)
{
   I->title = ezp_text(title, gdFontGiant, color->index);
}

EZP_Graph ezp_create_graph(EZP_Image I, int xtype, int ytype)
{
   EZP_Graph g = (EZP_Graph) malloc(sizeof(EZP_Graph_));
   EZP_Graph lg = I->graph;

   memset(g, 0, sizeof(EZP_Graph_));
   for (lg=I->graph;lg&&lg->next;lg=lg->next);
   if (lg) lg->next = g;
   else I->graph = g;

   g->next = NULL;
   I->ngraph++;
   g->image = (EZP_Image) I;
   g->plot = NULL;

   g->xo = 0;
   g->yo = 0;
   g->width = 0;
   g->height = 0;
 
   g->xtype = xtype;
   g->ytype = ytype;
  
   g->title = NULL;
   g->xcap = NULL;
   g->ycap = NULL;

   g->xset.f = (-1.0);
   g->yset.f = (-1.0);
   g->wset.f = (-1.0);
   g->hset.f = (-1.0);

   g->xlcb = NULL;
   g->ylcb = NULL;
   g->lcbarg = NULL;
   g->x_pad = 1;
   g->y_pad = 1;
   return(g);
}


void ezp_position_graph(EZP_Graph g, double x, double y, double w, double h)
{
   g->xset.f = x;
   g->yset.f = y;
   g->wset.f = w;
   g->hset.f = h;
}

void ezp_graph_captions(EZP_Graph G, char *xcap, char *ycap, EZP_Color color)
{
   if (xcap) G->xcap = ezp_text(xcap, gdFontLarge, color->index);
   if (ycap) G->ycap = ezp_text(ycap, gdFontLarge, color->index);
}


void ezp_set_graph_label_cb(EZP_Graph g,
       char *(*xcb)(V v, void *a), char *(*ycb)(V v, void *a), void *arg)
{
   g->xlcb = xcb;
   g->ylcb = ycb;
   g->lcbarg = arg;
}

EZP_Color ezp_allocate_color(EZP_Image I, char *color)
{
   EZP_Color c;
   RGB z;
   z = find_rgb(color);
   if (!z) return (NULL);
   c = (EZP_Color) malloc(sizeof(EZP_Color_));
   c->r = z->r;
   c->g = z->g;
   c->b = z->b;
   c->index = gdImageColorAllocate(I->gdimg, c->r, c->g, c->b);
   PRINTF("color %s (%d,%d,%d) is %d\n", color,c->r,c->g,c->b,c->index);
   return (c);
}

EZP_Color ezp_allocate_rgbcolor(EZP_Image I, int r, int g, int b)
{
   EZP_Color c;

   c = (EZP_Color) malloc(sizeof(EZP_Color_));
   c->r = r;
   c->g = g;
   c->b = b;
   c->index = gdImageColorAllocate(I->gdimg, r, g, b);
   PRINTF("color %d/%d/%d is %d\n", r,g,b,c->index);
   return (c);
}



EZP_Style ezp_create_style(int type, int d1, int d2,
         EZP_Color color)
{
   EZP_Style m;
   m = (EZP_Style) malloc(sizeof(EZP_Style_));
   m->type = type;
   m->d1 = d1;
   m->d2 = d2;
   if (m->type==EZP_STYLE_LINE) m->d1 = m->d2 = 1; /* compatibility */
   m->ci = color->index;
   return (m);
}

/* Plot xy data points.  For now just record the data. */

int ezp_plot_xy(EZP_Graph g, V *x, V *y, int np,
       EZP_Style m, char *legend)
{
   EZP_Plot P;
   int p;
   int ix, iy;

   P = (EZP_Plot) malloc(sizeof(EZP_Plot_));
   P->next = g->plot;
   g->plot = P;
   P->x = (V*) malloc(sizeof(V)*np);
   P->y = (V*) malloc(sizeof(V)*np);
   memcpy(P->x, x, sizeof(V)*np);
   memcpy(P->y, y, sizeof(V)*np);
   P->npoints = np;
   P->style = (EZP_Style) malloc(sizeof(EZP_Style_));
   memcpy(P->style, m, sizeof(EZP_Style_));

   if (legend) {
       EZP_Legend l, li;
       EZP_Image I = (EZP_Image)g->image;
       l = (EZP_Legend) malloc(sizeof(EZP_Legend_));
       l->text = ezp_text(legend, gdFontSmall, m->ci);
       l->style = (EZP_Style) malloc(sizeof(EZP_Style_));
       l->next = NULL;
       memcpy(l->style, m, sizeof(EZP_Style_));
       if (li=I->legend) {
          while (li->next) li=li->next;
          li->next = l;
       } else I->legend = l;
   }

   return (1);
}



/* Update min and max for v */
 
static void setminmax(int type, V v, V *min, V *max)
{
   switch (type) {
      case EZP_TYPE_INT:
      case EZP_TYPE_TIME:
          if (min) if (v.i < min->i) min->i = v.i;
          if (max) if (v.i > max->i) max->i = v.i;
          break;
      case EZP_TYPE_FLOAT:
          if (min) if (v.f < min->f) min->f = v.f;
          if (max) if (v.f > max->f) max->f = v.f;
          break;
   }
}

/* Get min and max values for an array of values */

static void get_range(int type, V *x, int np, V *min, V *max)
{
   int i;

   *min = *x;
   *max = *x;
   for (i=0;i<np; i++,x++) {
      setminmax(type, *x, min, max);
   }
}
 

/* Compute scale factors.

   Adjusts min/max to round numbers.  
   Roundness of numbers depends on their type.
   Returns the scale factor.

*/


/* adjustment for the start of the year in 1969 */

static int time_base(time_t t)
{
   int b;
   struct tm *tm;
   tm = localtime(&t);
   b = timezone;
   if (tm->tm_isdst>0) b -= HOUR;
   return (b);
}
   
   

/* return n where 'n<i' and 'n mod m = 0' */
static int64 mod_lower(int64 i, int64 m)
{
   int64 n = 0;
   /* PRINTF("  mod_lower: i=%lld, m=%lld ", i, m); */
   if (m<=0) return (i);
   if (i>0) n = (i/m) * m;
   if (i<0) n = ((i-m+1)/m) * m;
   /* PRINTF("  n=%lld\n", n); */
   return (n);
}
   
/* return n where 'n>i' and 'n mod m = 0' */
static int64 mod_upper(int64 i, int64 m)
{
   int64 n = 0;
   /* PRINTF("  mod_upper: i=%lld, m=%lld ", i, m); */
   if (m<=0) return (i);
   if (i>0) n = ((i+m-1)/m) * m;
   if (i<0) n = (i/m) * m;
   /* PRINTF("  n=%lld\n", n); */
   return (n);
}
   

/* return n where 'n<f' and 'n mod m = 0' */
static double mod_lower_f(double f, double m)
{
   double n = 0;
   double fm = fmod(f,m);
   PRINTF("  mod_lower_f: f=%g, m=%g ", f, m);
   if (m<=0) return (f);
   if (f>0) n = f - fm;
   if (f<0) {
      n = f - fm;
      if (fm!=0.0) n -= m;
   }
   PRINTF("  n=%g\n", n);
   return (n);
}
   
/* return n where 'n>f' and 'n mod m = 0' */
static double mod_upper_f(double f, double m)
{
   double n = 0;
   double fm = fmod(f,m);
   PRINTF("  mod_upper_f: f=%g, m=%g ", f, m);
   if (m<=0) return (f);
   if (f>0) {
      n = f - fm;
      if (fm!=0.0) n += m;
   }
   if (f<0) n = f - fm;
   PRINTF("  n=%g\n", n);
   return (n);
}

/* return the first day of the month/year for date = d */
/* note the time is offset */
static int64 mod_lower_d(int64 d, int64 m, int year)
{
   int base = time_base(d);
   time_t t = d + base;
   struct tm *lt;

   PRINTF("  mod_lower_d: i=%lld ", d);
   if (m<=0) return (d);
   lt = localtime(&t);
   lt->tm_sec = 0;
   lt->tm_min = 0;
   lt->tm_hour = 0;
   lt->tm_mday = 1;
   if (year) lt->tm_mon = 0;
   t = mktime(lt);
   PRINTF("  som=%d\n", t);
   t -= base;
   return ((int64)t);
}
   
/* return the last day of the month/year for date = d */
/* note the time is offset */
static int64 mod_upper_d(int64 d, int64 m, int year)
{
   int base = time_base(d);
   time_t t = d + base;
   struct tm *lt;

   PRINTF("  mod_upper_d: i=%lld ", d);
   if (m<=0) return (d);
   lt = localtime(&t);
   lt->tm_sec = 0;
   lt->tm_min = 0;
   lt->tm_hour = 0;
   lt->tm_mday = 1;
   if (year) {
      lt->tm_mon = 0;
      lt->tm_year++;
   } else {
      lt->tm_mon += 1;
      if (lt->tm_mon>11) {
         lt->tm_mon = 0;
         lt->tm_year++;
      }
   }
   t = mktime(lt);
   PRINTF("  eom=%d\n", t);
   t -= base;
   return ((int64)(t-1));
}
   
/* return # sec in months from d to d+w rounded up */
static int64 month_sec(int64 d, int64 w)
{
   time_t t = d;
   time_t ot;
   struct tm *lt;

   PRINTF("  month_sec: i=%lld ", d);
   while (w>=15*DAY) {
      ot = t;
      lt = localtime(&t);
      lt->tm_sec = 0;
      lt->tm_min = 0;
      lt->tm_hour = 0;
      lt->tm_mday = 1;
      lt->tm_mon += 1;
      if (lt->tm_mon>11) {
         lt->tm_mon = 0;
         lt->tm_year++;
      }
      t = mktime(lt);
      PRINTF(" m=%d ", t);
      w = w - ( t - ot );
   }
   PRINTF("  eom=%d\n", t);
   return ((int64)t-d+1);
}
   

   
int ipow(int64 base, int e)
{
   int64 p=base;
   if (e==0) return (1);
   for (;e>1;e--) p = p * base;
   return (p);
}

static int first_digit(int64 iv, int *mv)
{
   int m = 0;
   int d;
   double v = iv;
   if (iv<10) d = iv;
   else {
      m = log10(v);
      d = iv / ipow((int64)10, m);
   }
   *mv = m;
   PRINTF("    first_digit: in=%lld, out=%d, pow=%d\n", iv, d, m);
   return (d);
}

static int first_digit_f(double v, int *mv)
{
   int m = 0;
   int d;
   double rm;
   if (v==0.0) d = 0;
   else {
      if (v>=1.0) {
         rm = log10(v);
         m = rm;
         d = v / pow((double)10.0, (double)m);
         if ((d*pow((double)10.0, (double)m))!=v) d++;
         *mv = m;
      } else {
         rm = log10(1.0/v);
         m = rm + 1;
         d = v * pow((double)10.0, (double)(m));
         if ((d*pow((double)10.0, (double)m))!=v) d++;
         *mv = 0 - m;
      }
   }
   PRINTF("    first_digit_f: in=%g, out=%d, pow=%d\n", v, d, *mv);
   return (d);
}



char *D1(time_t d)
{
  static char str[48];
  char *e;
  strcpy(str, ctime(&d));
  if (e=strchr(str,'\n')) *e = '\0';
  return (str);
}
char *D2(time_t d)
{
  static char str[48];
  char *e;
  strcpy(str, ctime(&d));
  if (e=strchr(str,'\n')) *e = '\0';
  return (str);
}

/* compute the scale factors */

static Scale_ compute_scale(int type, V *min, V *max, int len, int pad)
{
   Scale_ scale;
   double fs;
   double fr, flen;
   int nltk /* # long ticks */;
   int nstk /* # short ticks */;
   V mn, mx;
   int ppt; /* pixels per tick */
   int i,j;
   int64 iscale = 0;
   double f;
   int64 ir, nir, imin, imax, ten_m;
   int64 ipmin = 0;
   int64 ipmax = 0;
   double r, nr, rmin, rmax;
   double rpmin = 0.0;
   double rpmax = 0.0;
   int64 bestdpt, bestnir=0;
   int64 bestmin, bestmax;
   double bestrpt, bestnr=0;
   double bestrmin, bestrmax;
   int bestnltk;
   int bestnstk;
   int fdpt, fdptm;
   int64 dpt, bdpt;
   int m;
   double rpt;
   int    frptm;
   int dt;
   char *d2;
   int use_nltk;
   int base;

   memset (&scale, 0, sizeof(Scale_));
   scale.tscale = 0;
   flen = len;
   switch (type) {
      case EZP_TYPE_INT:
         ir = max->i - min->i;
         if (ir==0) ir = 1;
         PRINTF("i_scale:   min=%lld, max=%lld\n", min->i, max->i);
         for (nltk=(len+D_TICKSPACE-1)/D_TICKSPACE;nltk>0;nltk--) {
            dpt = ir/nltk;
            if (dpt<1) dpt = 1;
            fdpt = first_digit(dpt, &fdptm);
            nstk = 1;
            if (fdpt==2) nstk = 2;
            if (fdpt>2) fdpt = mod_upper(fdpt,(int64)5), nstk=5; 
            ten_m = ipow(10, fdptm);
            dpt = fdpt * ten_m;
            imin = mod_lower(min->i, dpt);
            imax = mod_upper(max->i, dpt);
            ipmin = ipmax = 0;
            nir = (imax - imin);
            if (!pad) {
               ipmin = imin;
               ipmax = imax;
               imin = min->i;
               imax = max->i;
               nir = ir;
            }

            PRINTF("  nltk=%d, min=%lld, max=%lld, dpt=%lld, nir=%lld\n",
                 nltk, imin, imax, dpt, nir);
            if (bestnir==0 || nir<bestnir) {
               bestnltk = nltk;
               bestnstk = nstk;
               bestdpt = dpt;
               bestnir = nir;
               bestmin = imin;
               bestmax = imax;
            }
         }
         scale.tw.i = bestdpt;
         if (scale.tw.i==0) scale.tw.i = 1;
         scale.nlt = bestnltk;
         scale.nst = bestnstk;
         if (bestnir>0) scale.ppu = flen/(double)bestnir;
         else scale.ppu = 1.0;
         min->i = bestmin;
         max->i = bestmax;
         if (!pad) scale.tickpad.i = imin - ipmin;
 
         /* compute tick scale factor ( multiples of 3 ) */
         for (ir=1000,i=3;i<1000;i++) {
            if ((scale.tw.i/ir)*ir != scale.tw.i) break;
            ir *= 10;
         }
         i = (--i/3)*3;
         if (i>2) scale.tscale = i;

         PRINTF(" out: nlt=%d, nst=%d, range=%lld,%lld, pad=%lld, tw=%lld, ppu=%g, ts=%d\n",
                 scale.nlt, scale.nst, imin, imax, scale.tickpad.i, scale.tw.i, scale.ppu, scale.tscale);
         
         break;

      case EZP_TYPE_TIME:
         base = time_base(min->i);
         max->i -= base;
         min->i -= base;
         ir = max->i - min->i;
         if (ir==0) ir = 1;
         PRINTF("t_scale: flen=%g, DAY=%d, MON=%d\n    min=%lld (%s)\n    max=%lld (%s)\n",
            flen, DAY, MON, min->i, D1(min->i), max->i, D2(max->i));

         for (nltk=(len+D_TICKSPACE-1)/D_TICKSPACE;nltk>3;nltk--) {
            dpt = ir/nltk;
            PRINTF(" nltk=%d,  in dpt=%lld\n", nltk, dpt);
            nstk = 1;
            dt = EZP_DATETYPE_SEC;
            if (dpt>2) {
               dpt = mod_upper(dpt,(int64)5);
               bdpt = 5;
               nstk=5; 
            }
            if (dpt>10) {
               dpt = mod_upper(dpt,(int64)30);
               bdpt = 30;
               nstk=6;  
            }
            if (dpt>=MIN) dt = EZP_DATETYPE_MIN;
            if (dpt>MIN) {
               dpt = mod_upper(dpt,(int64)5*MIN);
               bdpt = 5*MIN;
               nstk=5; 
            }
            if (dpt>10*MIN) {
               dpt = mod_upper(dpt,(int64)30*MIN);
               bdpt = 30*MIN;
               nstk=6; 
            }
            if (dpt>30*MIN) {
               dpt = mod_upper(dpt,(int64)HOUR);
               bdpt = HOUR;
               nstk = dpt/(int64)HOUR;
               if (nstk<3) nstk=4; 
            }
            if (dpt>=HOUR) dt = EZP_DATETYPE_HOUR;

            if (dpt>=6*HOUR) {
               dpt = mod_upper(dpt,(int64)12*HOUR);
               bdpt = 12*HOUR;
               nstk=dpt/((int64)12*HOUR); 
               if (nstk<3) nstk = 4;
            }
            if (dpt>12*HOUR) {
               dpt = mod_upper(dpt,(int64)DAY);
               bdpt = DAY;
               nstk=dpt/(int64)DAY; 
               if (nstk<3) nstk = 4;
            }
            if (dpt>=DAY) dt = EZP_DATETYPE_DAY;

            if (dpt>=15*DAY || ir>3*MON) {
               dpt = mod_upper(dpt,(int64)MON);
               bdpt = MON;
               nstk=dpt/(MON); 
               if (nstk<2) nstk=2;
            }
            if (dpt>=MON) dt = EZP_DATETYPE_MON;

            if (dpt>=6*MON || ir>3*YEAR) {
               dpt = mod_upper(dpt,(int64)YEAR);
               bdpt = YEAR;
               nstk=dpt/(YEAR/NYST); 
               if (nstk<2) nstk=2;
            }
            if (dpt>=YEAR) dt = EZP_DATETYPE_YEAR;

            if (dt == EZP_DATETYPE_YEAR) {
               imin = mod_lower_d(min->i, bdpt, 1);
               imax = mod_upper_d(max->i, dpt, 1);
            } else if (dt == EZP_DATETYPE_MON) {
               imin = mod_lower_d(min->i, bdpt, 0);
               imax = mod_upper_d(max->i, dpt, 0);
            } else {
               imin = mod_lower(min->i, bdpt);
               imax = mod_upper(max->i, dpt);
            }

            if (!pad) {
               ipmin = imin;
               ipmax = imax;
               imin = min->i;
               imax = max->i;
               nir = ir;
            } else {
               nir = dpt*nltk;
            }


            use_nltk = nltk;
            if (imin+nir<max->i) {
               use_nltk++;
               nir = dpt*use_nltk;
            }


            PRINTF("     min=%lld (%s)\n     max=%lld (%s)\n     dpt=%lld, nir=%lld, nltk=%d, nstk=%d\n",
                 imin, D1(imin), imax, D2(imax), dpt, nir, use_nltk, nstk);

            if (bestnir==0 || nir<bestnir) {
               bestnir = nir;
               scale.nlt = use_nltk;
               scale.nst = nstk;
               if (nir>0) scale.ppu = flen/(double)nir;
               else scale.ppu = 1.0;
               scale.tw.i = dpt;
               if (scale.tw.i==0) scale.tw.i = 1;
               scale.min.i = imin;
               scale.max.i = imax;
               if (!pad) scale.tickpad.i = imin - ipmin;
               scale.dt = dt;
               PRINTF("   ** best so far\n");
            }
         }
         PRINTF("   out: nlt=%d, nst=%d, range=%lld,%lld, pad=%lld, tw=%lld, ppu=%g\n\n",
                 scale.nlt, scale.nst, imin, imax, scale.tickpad.i, scale.tw.i, scale.ppu);
         min->i = scale.min.i + base;
         max->i = scale.max.i + base;
         break;


     case EZP_TYPE_FLOAT:
         r = max->f - min->f;
         if (r==0) r = 1.0;
         PRINTF("f_scale:   min=%g, max=%g\n", min->f, max->f);
         for (nltk=(len+D_TICKSPACE-1)/D_TICKSPACE;nltk>1;nltk--) {
            rpt = r/nltk;
            fdpt = first_digit_f(rpt, &fdptm);
            nstk = 1;
            if (fdpt==2) nstk = 2;
            if (fdpt>2) fdpt = mod_upper(fdpt,(int64)5), nstk=5;
            rpt = ((double) fdpt) * pow((double)10.0, (double)fdptm);
            rmin = mod_lower_f(min->f, rpt);
            rmax = mod_upper_f(max->f, rpt);
            nr = (rmax - rmin);
            if (!pad) {
               rpmin = rmin;
               rpmax = rmax;
               rmin = min->f;
               rmax = max->f;
               nr = r;
            }
            PRINTF("  nltk=%d, min=%g, max=%g, dpt=%g, nr=%g\n",
                 nltk, rmin, rmax, rpt, nr);
            if (bestnr==0 || nr<bestnr) {
               bestnltk = nltk;
               bestnstk = nstk;
               bestrpt = rpt;
               bestnr = nr;
               bestrmin = rmin;
               bestrmax = rmax;
            }
         }
         scale.tw.f = bestrpt;
         scale.nlt = bestnltk;
         scale.nst = bestnstk;
         if (bestnr>0) scale.ppu = flen/bestnr;
         else scale.ppu = 1.0;
         if (!pad) scale.tickpad.f = rmin - rpmin;
         PRINTF(" out: nlt=%d, nst=%d, range=%g,%g, pad=%g, tw=%g, ppu=%g\n",
                 scale.nlt, scale.nst, rmin, rmax, scale.tickpad.f, bestrpt, scale.ppu);
         min->f = bestrmin;
         max->f = bestrmax;
         break;


   }
   return (scale);
}


/* Format the tick value */


static char *month_txt[] = {
  "Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};

static char *tickstring(int type, V tick, Scale scale)
{
   static char pbuf[128];
   struct tm *tkt;
   time_t ttime;
   int fpc;
   char ffmt[24];
   double f;
   int64 ts;
   double fts;
   int i;

   for (ts=1,i=0;i<scale->tscale;i++) ts = ts * 10;
   fts = scale->tscale;
   fts = pow(10.0,fts);
   switch (type) {
     case EZP_TYPE_INT:
         sprintf(pbuf,"%lld", tick.i/ts);
         break;
     case EZP_TYPE_FLOAT:
         sprintf(pbuf,"%g", tick.f/fts);
         break;
     case EZP_TYPE_TIME:
         ttime = tick.i;
         if (scale->dt>=EZP_DATETYPE_MON) ttime += 2*HOUR; /* allow for DST */
         tkt = localtime(&ttime);
         if (scale->dt==EZP_DATETYPE_SEC) sprintf(pbuf,"%02.2d:%02.2d",
                tkt->tm_min, tkt->tm_sec);
         else if (scale->dt==EZP_DATETYPE_MIN) sprintf(pbuf,"%02.2d:%02.2d",
                tkt->tm_hour, tkt->tm_min);
         else if (scale->dt==EZP_DATETYPE_HOUR) sprintf(pbuf,"%02.2d:%02.2d",
                tkt->tm_hour, tkt->tm_min);
         else sprintf(pbuf,"%s %02.2d", month_txt[tkt->tm_mon], tkt->tm_mday);
   }
   return (pbuf);
}


static char *tickstringsup(int type, V tick, Scale scale, int dt, int first)
{
   static char pbuf[128];
   struct tm *tkt;
   time_t ttime;
   int fpc;
   char ffmt[24];
   double f;
   static int ly;
   static int lm;

   pbuf[0] = '\0';
   switch (type) {
     case EZP_TYPE_INT:
         break;
     case EZP_TYPE_FLOAT:
         break;
     case EZP_TYPE_TIME:
         ttime = tick.i;
         if (scale->dt>=EZP_DATETYPE_MON) ttime += 2*HOUR; /* allow for DST */
         tkt = localtime(&ttime);

         if ((!first) && (tkt->tm_mday==lm && tkt->tm_year==ly)) break;
         if (dt>=EZP_DATETYPE_DAY) {
           if (first || tkt->tm_year!=ly) sprintf(pbuf,"%d", tkt->tm_year+1900);
         } else {
            if (first) sprintf(pbuf,"%s %d, %d",
                  month_txt[tkt->tm_mon], tkt->tm_mday,
                  tkt->tm_year+1900);
            else if (tkt->tm_mday!=lm || tkt->tm_year!=ly) 
                  sprintf(pbuf,"%s %d", month_txt[tkt->tm_mon], tkt->tm_mday);
         }
         lm = tkt->tm_mday;
         ly = tkt->tm_year;
   }
   return (pbuf);
}

int ezp_make_plots(EZP_Image I)
{
   EZP_Plot p;
   EZP_Graph g;
   V xmin, xmax;
   V ymin, ymax;
   int first;
   int dw, dh, gw, gh;
   int sxo;
   int sx,ex,sy,ey;
   Scale_ xscale, yscale;
   int i;
   V *x;
   V *y;
   int ix, iy;
   int ixe, ixse, iye;
   V vx, vy;
   V tick;
   EZP_Legend L;

   int lg_height = 0;
   int lg_maxw = 0;
   int nlg = 0;
   int nlgcol;
   int lgcol;
   int lgline;
   int firsttick;
   int64 tw;

   int gno;
   double fpad;
   int ipad;

   double zero = 0.0;
   
   /* get min and max for x and y values for each graph */

   infinity = (double)1.0/zero;
   for (g=I->graph;g;g=g->next) {
     first = 1;
     for (p=g->plot;p;p=p->next) {
        get_range(g->xtype, p->x, p->npoints, &xmin, &xmax);
        get_range(g->ytype, p->y, p->npoints, &ymin, &ymax);
        if ( (g->xtype==EZP_TYPE_FLOAT && 
               (xmin.f==-infinity || xmax.f==infinity)) ||
             (g->ytype==EZP_TYPE_FLOAT && 
               (ymin.f==-infinity || ymax.f==infinity)) ) {
           fprintf(stderr, "ezp cannot plot infinite values!\n");
           return (0);
        }
        if (first) {
           g->xmin = xmin;
           g->xmax = xmax;
           g->ymin = ymin;
           g->ymax = ymax;
           first = 0;
        } else {
           setminmax(g->xtype, xmin, &g->xmin, NULL);
           setminmax(g->xtype, xmax, NULL, &g->xmax);
           setminmax(g->ytype, ymin, &g->ymin, NULL);
           setminmax(g->ytype, ymax, NULL, &g->ymax);
        }
        g->nplot++;
     }
   }
   
   /* Format and print the legend  */

   for (L=I->legend;L;L=L->next) {
      lg_height += L->text->height;
      if (L->text->width > lg_maxw) lg_maxw = L->text->width;
      nlg++;
   }
   if (lg_height) {
  
     nlgcol = (I->width-D_BORDER) / (lg_maxw+20+D_BORDER);
     if (nlgcol==0) nlgcol = 1;
     if (nlgcol>nlg) nlgcol = nlg;
     PRINTF("Formatting %d legends in %d columns\n", nlg, nlgcol);
     if (nlgcol>1) lg_height = (lg_height+nlgcol-1)/nlgcol;
     nlgcol = (nlg+nlgcol-1)/nlgcol; /* number lines per column */
     lg_height += D_BORDER;
     gdImageLine(I->gdimg, D_BORDER, Y(lg_height+4),
         I->width-D_BORDER, Y(lg_height+4), 1);

     ix = D_BORDER;
     iy = lg_height;
     lgcol = 0;
     lgline = 0;
     for (L=I->legend;L;L=L->next) {
       EZP_Style S;
       S = L->style;
       switch (S->type) {
         case EZP_STYLE_BOX:
            gdImageFilledRectangle(I->gdimg,
               ix, Y(iy),
               ix+(S->d1), Y(iy-(S->d2)), S->ci);
            break;
         case EZP_STYLE_OPEN_BOX:
            gdImageRectangle(I->gdimg,
               ix, Y(iy),
               ix+(S->d1), Y(iy-(S->d2)), S->ci);
            break;
         case EZP_STYLE_CIRCLE:
         case EZP_STYLE_OPEN_CIRCLE:
            gdImageArc(I->gdimg,
               ix, Y(iy-5),
               S->d1, S->d1, 0, 360, S->ci);
            if (S->type==EZP_STYLE_CIRCLE) gdImageFillToBorder(
                      I->gdimg, ix, Y(iy-5), S->ci, S->ci);
            break;
         case EZP_STYLE_LINE:
         case EZP_STYLE_FAT_LINE:
            PRINTF("-- setting legend line thickness to %d\n", S->d2);
            gdImageSetThickness(I->gdimg, S->d2);
            gdImageLine(I->gdimg,
               ix, Y(iy-7),
               ix+10, Y(iy-7), S->ci);
            gdImageSetThickness(I->gdimg, 1);
            break;
         case EZP_STYLE_H_BAR:
         case EZP_STYLE_V_BAR:
            gdImageFilledRectangle(I->gdimg,
               ix, Y(iy),
               ix+(S->d1), Y(iy-(S->d1)), S->ci);
            break;
       }
       gdImageString(I->gdimg, gdFontSmall, ix+20, Y(iy),
               UC L->text->str, 1);
       iy -= L->text->height;
       if (++lgline == nlgcol) {
          lgcol++;
          ix = D_BORDER + lgcol*(lg_maxw+20+D_BORDER);
          iy = lg_height;
          lgline = 0;
       }
     }
   }


   /* Draw each graph on the image */

   for (gno=0,g=I->graph;g;g=g->next,gno++) {
      double gx = g->xset.f;
      double gy = g->yset.f;
      double gw = g->wset.f;
      double gh = g->hset.f;
      int ixo, iyo;
    
      if (gx<0) gx = gno*(1.0/I->ngraph);
      if (gy<0) gy = 0.0;
      if (gw<0) gw = 1.0 / I->ngraph;
      if (gh<0) gh = 1.0;

      ixo = g->ycap->height + D_CAPSEP + D_TICKLEN; 
      if (!g->xo) g->xo = D_BORDER + ixo + (gx*(I->width-2*D_BORDER));

      iyo = g->xcap->height + D_CAPSEP + D_TICKLEN; 
      /* dates may have a year in the label */
      if (g->xtype == EZP_TYPE_TIME) iyo += D_TICKLEN;
      if (!g->yo) g->yo = D_BORDER+lg_height+iyo+(gy*(I->height-2*D_BORDER));

      dw = (I->width)*gw - ixo - 3*D_BORDER;
      dh = (I->height-lg_height)*gh - iyo -
                         I->title->height - D_CAPSEP - 2*D_BORDER;

      if (!g->width) g->width = dw;
      if (!g->height) g->height = dh;

      PRINTF("set graph xo/yo = %d/%d, w/h = %d/%d\n",
               g->xo, g->yo, g->width, g->height);

      /* Compute the scale factors for this graph */

      g->xscale = compute_scale(g->xtype,
              &g->xmin, &g->xmax, g->width, g->x_pad);
      g->yscale = compute_scale(g->ytype,
              &g->ymin, &g->ymax, g->height, g->y_pad);
      

      /* Draw graph axes */
   
      sx = g->xo;
      sy = g->yo;
      ex = sx + g->width;
      ey = sy + g->height;
      gdImageLine(I->gdimg, sx, Y(sy), ex, Y(sy), 1);
      gdImageLine(I->gdimg, sx, Y(sy), sx, Y(ey), 1);
      
      /* Plot data points (or lines) */

      for (p=g->plot;p;p=p->next) {
         EZP_Style S = p->style;
         int pix = 0;
         int piy = 0;
         int ix0, ix1, iy0, iy1;

         PRINTF("-- setting data line thickness to %d\n", S->d1);
         gdImageSetThickness(I->gdimg, S->d1); 

         for (i=0,x=p->x,y=p->y;i<p->npoints;i++,x++,y++) {
            switch (g->xtype) {
               case EZP_TYPE_INT:
               case EZP_TYPE_TIME:
                 ix = g->xo + (x->i - g->xmin.i) * g->xscale.ppu;
                 break;
               case EZP_TYPE_FLOAT:
                 ix = g->xo + (x->f - g->xmin.f) * g->xscale.ppu;
            }
            switch (g->ytype) {
               case EZP_TYPE_INT:
               case EZP_TYPE_TIME:
                 iy = g->yo + (y->i - g->ymin.i) * g->yscale.ppu;
                    break;
               case EZP_TYPE_FLOAT:
                 iy = g->yo + (y->f - g->ymin.f) * g->yscale.ppu;
            }
   
            switch (S->type) {
               case EZP_STYLE_BOX:
                  /* PRINTF(" box at (%d,%d)\n", ix, iy); */
                  gdImageFilledRectangle(((EZP_Image)(g->image))->gdimg,
                         ix-S->d1/2, Y(iy+S->d2/2),
                           ix+S->d1/2, Y(iy-S->d2/2), S->ci);
                  break;
               case EZP_STYLE_OPEN_BOX:
                  /* PRINTF(" obox at (%d,%d)\n", ix, iy); */
                  gdImageRectangle(((EZP_Image)(g->image))->gdimg,
                         ix-S->d1/2, Y(iy+S->d2/2),
                           ix+S->d1/2, Y(iy-S->d2/2), S->ci);
                  break;
               case EZP_STYLE_CIRCLE:
               case EZP_STYLE_OPEN_CIRCLE:
                  gdImageArc(I->gdimg, ix, Y(iy),
                     S->d1, S->d1, 0, 360, S->ci);
                  if (S->type==EZP_STYLE_CIRCLE) gdImageFillToBorder(
                            I->gdimg, ix, Y(iy), S->ci, S->ci);
                  break;
               case EZP_STYLE_LINE:
               case EZP_STYLE_FAT_LINE:
                  /* PRINTF(" line at (%d,%d)\n", ix, iy); */
                  if (pix) gdImageLine(((EZP_Image)(g->image))->gdimg,
                         pix, Y(piy), ix, Y(iy), S->ci);
                  pix = ix;
                  piy = iy;
                  break;
               case EZP_STYLE_V_BAR:
                  /* PRINTF(" vbar at (%d,%d)\n", ix, iy); */
                  gdImageFilledRectangle(((EZP_Image)(g->image))->gdimg,
                         ix-S->d1/2, Y(iy),
                           ix+S->d1/2, Y(g->yo), S->ci);
                  break;
               case EZP_STYLE_H_BAR:
                  /* PRINTF(" hbar at (%d,%d)\n", ix, iy); */
                  gdImageFilledRectangle(((EZP_Image)(g->image))->gdimg,
                         g->xo, Y(iy+S->d2/2),
                           ix, Y(iy-S->d2/2), S->ci);
                  break;
               
            }
         }
      }
      gdImageSetThickness(I->gdimg, 1); 
    
      /* Write tick marks and labels */
   
      if (g->nplot>0) {

         /* X axis */

         ix = g->xo;
         ixe = 0;
         ixse = 0;
         iy = g->yo;
         iye = 0;
         tick = g->xmin;
         if (g->xtype==EZP_TYPE_FLOAT) {
            tick.f -= g->xscale.tickpad.f;
            fpad = (g->xscale.tickpad.f)*(g->xscale.ppu);
            ipad = fpad;
         } else {
            tick.i -= g->xscale.tickpad.i;
            fpad = (g->xscale.tickpad.i*1.0)*(g->xscale.ppu);
            ipad = fpad;
         }
         firsttick = 1;
         if (g->xscale.tscale && !g->xlcb) {
            char tstxt[24];
            sprintf(tstxt,"x10^%d", g->xscale.tscale);
            gdImageString(I->gdimg, gdFontSmall,
                  ix, Y(iy-3*D_TICKLEN+5), UC tstxt, 1);
         }
 
   PRINTF(" ofsetting x by %d (%lld)\n", ipad, g->xscale.tickpad.i);
          ix -= ipad;
          vx.f = ix;
         gdImageSetThickness(I->gdimg, 1); 
         while (ix<=g->xo+g->width) {
            char *ttxt;
            int ttxtl, stxtl;
            double dxst;
            double fis;
            long int is;
   
   PRINTF(" .. try x/xo = %d/%d\n", ix, g->xo);
            if (g->xlcb) ttxt = g->xlcb(tick, g->lcbarg);
            else ttxt = tickstring(g->xtype, tick, &g->xscale);
            ttxtl = strlen(ttxt) * gdFontSmall->w;
            if (ix>=g->xo) gdImageLine(I->gdimg, ix, Y(iy), ix, Y(iy-D_TICKLEN), 1);
            if (ix>ixe+5) {
              int ixts = ix - ttxtl/2;
              if (ix>=g->xo) gdImageString(I->gdimg, gdFontSmall, ixts,
                   Y(iy-D_TICKLEN),
                   UC ttxt, 1);
              ixe = ix + ttxtl;
   
              if (firsttick || (g->xscale.dt >= EZP_DATETYPE_HOUR)) {
                 ttxt = tickstringsup(g->xtype, tick, &g->xscale, g->xscale.dt,
                         firsttick);
                 stxtl = strlen(ttxt) * gdFontSmall->w;
                 if (ix>ixse+5) {
                   int ixts = ix - stxtl/2;
                   if (ix>=g->xo) gdImageString(I->gdimg, gdFontSmall, ixts,
                       Y(iy-2*D_TICKLEN),
                       UC ttxt, 1);
                   ixse = ix + stxtl;
                 }
                 firsttick = 0;
              }
            }
   
            /* Draw short ticks */
            if (g->xscale.nst) dxst = (g->xscale.tw.i * g->xscale.ppu) / 
                     g->xscale.nst;
            fis = ix;
            for (i=1;i<g->xscale.nst;i++) {
               if (g->xtype==EZP_TYPE_TIME && g->xscale.dt>=EZP_DATETYPE_MON) {
                  int nm = 1;
                  is = fis;
                  if (g->xscale.dt==EZP_DATETYPE_YEAR) nm = (12/NYST);
                  fis += (month_sec(is,MON*nm)*g->xscale.ppu);
               } else fis += dxst;
               is = fis;
               if (is > g->xo+g->width) break;
               if (is>=g->xo) gdImageLine(I->gdimg, is, Y(iy), is,
                   Y(iy-D_TICKLEN/2), 1);
            }
   
            /* Next labelled tick */
            switch (g->xtype) {
              case EZP_TYPE_INT:
                vx.f += g->xscale.tw.i * g->xscale.ppu; 
                ix = vx.f;
                tick.i += g->xscale.tw.i;
                break;
              case EZP_TYPE_TIME:
                tw = g->xscale.tw.i;
                if (g->xscale.dt>=EZP_DATETYPE_MON) tw = month_sec(tick.i,tw);
                vx.f += tw * g->xscale.ppu; 
                ix = vx.f;
                tick.i += tw;
                break;
              case EZP_TYPE_FLOAT:
                vx.f += g->xscale.tw.f * g->xscale.ppu; 
                ix = vx.f;
                tick.f += g->xscale.tw.f;
            }
         }
         
         /* Y axis */

         ix = g->xo;
         iy = g->yo;
         tick = g->ymin;
         if (g->ytype==EZP_TYPE_FLOAT) {
            tick.f -= g->yscale.tickpad.f;
            fpad = (g->yscale.tickpad.f)*(g->yscale.ppu);
            ipad = fpad;
         } else {
            tick.i -= g->yscale.tickpad.i;
            fpad = (g->yscale.tickpad.i*1.0)*(g->yscale.ppu);
            ipad = fpad;
         }
         if (g->yscale.tscale) {
            char tstxt[24];
            sprintf(tstxt,"x10^%d", g->yscale.tscale);
            gdImageString(I->gdimg, gdFontSmall,
                             ix - D_TICKLEN,
                             D_BORDER + I->title->height, UC tstxt, 1);
         }
   PRINTF(" ofsetting y by %d (%lld)\n", ipad, g->yscale.tickpad.i);
         iy -= ipad;
         vy.f = iy;
         while (iy<=g->yo+g->height) {
            char *ttxt;
            int ttxtl;
            double dxst;
            int is;
   
            if (g->ylcb) ttxt = g->ylcb(tick, g->lcbarg);
            else ttxt = tickstring(g->ytype, tick, &g->yscale);
            ttxtl = strlen(ttxt) * gdFontSmall->w;
            if (iy>=g->yo) gdImageLine(I->gdimg, ix-D_TICKLEN, Y(iy), ix, Y(iy), 1);
            if (iy>iye+5) {
               if (iy>=g->yo) gdImageStringUp(I->gdimg, gdFontSmall,
                   ix-D_TICKLEN-gdFontSmall->h-3, Y(iy-ttxtl/3),
                   UC ttxt, 1);
              iye = iy + ttxtl;
            }
   
            /* Draw short ticks */
            if (g->yscale.nst) dxst = (g->yscale.tw.f * g->yscale.ppu) / 
                     g->yscale.nst;
            for (i=1;i<g->yscale.nst;i++) {
               is = iy + dxst*i;
               if (is > g->yo+g->height) break;
               if (is>g->yo) gdImageLine(I->gdimg, ix-D_TICKLEN/2, Y(is),
                   ix, Y(is), 1);
            }
   
            if (g->ymin.i==g->ymax.i) break;
   
            /* Next labelled tick */
            switch (g->ytype) {
              case EZP_TYPE_INT:
                vy.f += g->yscale.tw.i * g->yscale.ppu;
                iy = vy.f;
                tick.i += g->yscale.tw.i;
                break;
              case EZP_TYPE_TIME:
                tw = g->yscale.tw.i;
                if (g->yscale.dt>=EZP_DATETYPE_MON) tw = month_sec(tick.i,tw);
                vy.f += tw * g->yscale.ppu;
                iy = vy.f;
                tick.i += tw;
                break;
              case EZP_TYPE_FLOAT:
                vy.f += g->yscale.tw.f * g->yscale.ppu;
                iy = vy.f;
                tick.f += g->yscale.tw.f;
            }
         }

      }
   
      /* write captions */
   
      if (g->xcap) {
         ix = g->xo;
         iy = D_BORDER + g->xcap->height + lg_height ;
         gdImageString(I->gdimg, g->xcap->font, ix, Y(iy),
                   UC g->xcap->str, g->xcap->ci);
      }
   
      if (g->ycap) {
         ix = g->xo - g->ycap->height*2 - D_TICKLEN;
         iy = g->yo;
         gdImageStringUp(I->gdimg, g->ycap->font, ix, Y(iy),
                   UC g->ycap->str, g->ycap->ci);
      }
   }

   /* write title */

   iy = D_BORDER;
   ix = (I->width - I->title->width)/2;
   gdImageString(I->gdimg, I->title->font, ix, iy,
             UC I->title->str, I->title->ci);
      
   I->isplotted = 1;

   return (1);
}


int ezp_write_image(EZP_Image img, int type, FILE *f)
{
   if (!img->isplotted) ezp_make_plots(img);

   if (!f) return (0);
   switch (type) {
      case EZP_IMAGE_JPG:
         gdImageJpeg(img->gdimg, f, 95);
         break;
      case EZP_IMAGE_PNG:
         gdImagePng(img->gdimg, f);
         break;
   }
   return (1);
}

int ezp_save_image(EZP_Image img, int type, char *file)
{
   FILE *f;
  
   f = fopen(file, "wb");
   if (!f) return (0);
   ezp_write_image(img, type, f);
   fclose(f);
   return (1);
}

/* Return the critical attributes of a graph. */

EZP_Graph_attrs ezp_get_graph_attrs(EZP_Graph g)
{
   EZP_Graph_attrs ga = (EZP_Graph_attrs)malloc(sizeof(EZP_Graph_attrs_));
   ga->xo = g->xmin;
   ga->yo = g->ymin;
   ga->xp = g->xo;
   ga->yp = g->yo;
   ga->xs = 1.0/g->xscale.ppu;
   ga->ys = 1.0/g->yscale.ppu;
   return (ga);
}

/* Set the padding flags */

void ezp_set_graph_pad_flags(EZP_Graph g, int x, int y)
{
   g->x_pad = x;
   g->y_pad = y;
}


