/* ========================================================================
 * Copyright (c) 2004-2006 The University of Washington
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ========================================================================
 */


/* EZPlot definitions */

#define D_BORDER     12
#define D_CAPSEP     16
#define D_TICKLEN    10
#define D_TICKSPACE  50

#if !defined(HAVE_INT64)
typedef long long int int64;
#endif

typedef union {
  int64 i;
  double f;
} V;

#define EZP_STYLE_BOX          11
#define EZP_STYLE_OPEN_BOX     12
#define EZP_STYLE_CIRCLE       21
#define EZP_STYLE_OPEN_CIRCLE  22
#define EZP_STYLE_LINE         31
#define EZP_STYLE_FAT_LINE     32

#define EZP_STYLE_H_BAR        41
#define EZP_STYLE_V_BAR        42

#define EZP_TYPE_INT            1
#define EZP_TYPE_FLOAT          2
#define EZP_TYPE_TIME           3

#define EZP_DATETYPE_SEC        1
#define EZP_DATETYPE_MIN        2
#define EZP_DATETYPE_HOUR       3
#define EZP_DATETYPE_DAY        4
#define EZP_DATETYPE_MON        5
#define EZP_DATETYPE_YEAR       6

typedef struct EZP_Color__ {
  char *name;
  int r;
  int g;
  int b;
  int index;
} EZP_Color_, *EZP_Color;


#define EZP_FONT_TINY      0
#define EZP_FONT_SMALL     1
#define EZP_FONT_MEDIUM    2
#define EZP_FONT_LARGE     3
#define EZP_FONT_GIANT     4

typedef struct EZP_Style__ {
  int type;
  int d1;
  int d2;
  int ci;
} EZP_Style_, *EZP_Style;

#ifdef EZP_LIBRARY
typedef struct EZP_Text__ {
  char *str;
  gdFontPtr font;
  int ci;
  int width;
  int height;
  int nl;
} EZP_Text_, *EZP_Text;
 
typedef struct EZP_Legend__ {
  struct EZP_Legend__ *next;
  EZP_Text text;
  EZP_Style style;
} EZP_Legend_, *EZP_Legend;

#endif


typedef struct Scale__ {
  double ppu;   /* pixels / unit */
  V     tw;     /* tick width (range / tick) */
  int   nlt;    /* number of long ticks */
  int   nst;    /* number of short ticks */
  int   dt;     /* date type */
  V base;
  V     min;
  V     max;
  int   tscale; /* scale factor for ticks (e.g. x1000) */
  V     tickpad; /* padding: tick to graph edge */
} Scale_, *Scale;

typedef struct EZP_Plot__ {
  struct EZP_Plot__ *next;
  int npoints;
  V *x;
  V *y;
  EZP_Style style;
} EZP_Plot_, *EZP_Plot;

typedef struct EZP_Graph__ {
  void *image;                 /* image containing this graph */
  struct EZP_Graph__ *next;
  EZP_Plot plot;               /* plots */
  int nplot;
  int xtype;                   /* data type of x axis */
  int ytype;                   /* data type of y axis */
                               /* Followig offsets are relative to the image */
  int xo;                      /* x offset (pixels) of x origin from left */
  int yo;                      /* y offset (pixels) of y origin from bottom */
  int width;                   /* width of the plotable area */
  int height;                  /* height of the plotable area */
  int weight;
  V xmin;                      /* x value at origin */
  V xmax;                      /* x value at graph edge */
  V ymin;                      /* y value at origin */
  V ymax;                      /* y value at graph top */
  int xminset;
  int xmaxset;
  int yminset;
  int ymaxset;
  Scale_ xscale;               /* x scale factors */
  Scale_ yscale;               /* y scale factors */
#ifdef EZP_LIBRARY
  EZP_Text title;
  EZP_Text xcap;
  EZP_Text ycap;
#else
  void *title;
  void *xcap;
  void *ycap;
#endif
  V xset;
  V yset;
  V wset;
  V hset;
  char *(*xlcb)(V v, void *a);   /* label callback (user labels) */
  char *(*ylcb)(V v, void *a);   
  void *lcbarg;                  /* arg for callback */
  int x_pad;                     /* true to pad border to tick */
  int y_pad;
} EZP_Graph_, *EZP_Graph;
  
typedef struct EZP_Image__ {
#ifdef EZP_LIBRARY
  gdImagePtr gdimg;
  EZP_Text title;
  EZP_Legend legend;
#else
  void *gdimg;
  void *title;
  void *legend;
#endif
  EZP_Graph graph;
  int ngraph;
  int width;
  int height;
  int isplotted;
} EZP_Image_, *EZP_Image;



#define EZP_IMAGE_JPG   1
#define EZP_IMAGE_PNG   2


extern int ezp_verbose;

/* Graph attrs for users.  Note the name changes. */

typedef struct EZP_Graph_attrs__ {
  V xo;        /* x value at origin */
  V yo;        /* y value at origin */
  int xp;      /* pixel position of x origin on plot */
  int yp;      /* pixel position of y origin on plot */
  double xs;   /* x scale (range/pixel) */
  double ys;   /* y scale (range/pixel) */
} EZP_Graph_attrs_, *EZP_Graph_attrs;


/* functions */

EZP_Image ezp_create_image(int width, int height);
EZP_Graph ezp_create_graph(EZP_Image I, int xtype, int ytype);
void ezp_position_graph(EZP_Graph g, double x, double y, double w, double h);
EZP_Color ezp_allocate_color(EZP_Image I, char *name);
EZP_Color ezp_allocate_rgbcolor(EZP_Image I, int r, int g, int b);
EZP_Style ezp_create_style(int type, int d1, int d2,
         EZP_Color color);
int ezp_plot_xy(EZP_Graph g, V *x, V *y, int np,
       EZP_Style m, char *legend);
int ezp_save_image(EZP_Image img, int type, char *file);
void ezp_set_graph_label_cb(EZP_Graph g,
    char *(*xcb)(V v, void *a), char *(*ycb)(V v, void *a), void *arg);

EZP_Graph_attrs ezp_get_graph_attrs(EZP_Graph g);
void ezp_set_graph_pad_flags(EZP_Graph g, int x_pad, int y_pad);


